/* vi: set sw=4 ts=4: */
/*
 * Gtk+2.x  Utility routines: Disk Manager window
 *
 * Copyright (C) 2002-2006 by Tito Ragusa <tito-wolit@tiscali.it>
 *
 * Based on Mini df implementation for busybox
 *
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * based on original code by (I think) Bruce Perens <bruce@pixar.com>.
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *
 * Mini mount implementation for busybox
 *
 * Copyright (C) 1995, 1996 by Bruce Perens <bruce@pixar.com>.
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005 by Rob Landley <rob@landley.net>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 *

 */

#include <limits.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <strings.h>
#include <stdio.h>
#include <mntent.h>
#include <ctype.h>
#include <sys/mount.h>
#include <sys/vfs.h>
#include <gtk/gtk.h>

/* These two aren't always defined in old headers */
#ifndef MS_BIND
#define MS_BIND		4096
#endif
#ifndef MS_MOVE
#define MS_MOVE		8192
#endif

typedef	struct	Fs_data			Fs_data;
struct Fs_data {
	const char *device;
	int *ptr_found;
};

static char *mount_path = NULL;
static char *is_mounted  = NULL;
static char *mount_opts = NULL;
static char *mount_device = NULL;
static char *mount_type = NULL;

gboolean tot_list_foreach_func (GtkTreeModel *model, GtkTreePath ATTRIBUTE_UNUSED *path, 
								GtkTreeIter *iter,   Fs_data *user_data)
{
	char *device;
	int ret = FALSE;
	struct Fs_data *F_data;

	F_data = user_data;

	/* Note: here we use 'iter' and not '&iter', because we did not allocate
	*  the iter on the stack and are already getting the pointer to a tree iter */
	gtk_tree_model_get (model, iter, 0 ,&device, -1);
	if (strcmp (F_data->device, device) == 0) {
		*F_data->ptr_found = 1;
		ret = TRUE;
	}
	return ret;
}

void df_list(GtkWidget *clist)
{
	char percent[5];
	char *size=NULL;
	char *used=NULL;
	char *avail=NULL;
	long blocks_used;
	long blocks_percent_used;
	FILE *mount_table;
	struct mntent *mount_entry;
	struct statfs s;
	struct Fs_data  *Fdata;
	char *bb_path_mtab_file	= "/etc/mtab";
	const char *row[9];
	int found = 0;

	Fdata = xmalloc(sizeof(Fs_data *));
SECOND_PASS:
	/* Check if reading the mtab file failed */
	if ((mount_table  = setmntent( bb_path_mtab_file , "r")) == NULL) {
		/* Bummer.  fall back on trying the /proc filesystem */
		if ((strcmp(bb_path_mtab_file, "/etc/fstab") == 0) 
			|| (mount_table = setmntent("/proc/mounts", "r")) == NULL) {
			bb_perror_msg("%s", bb_path_mtab_file);
			return;
		} else {
			bb_path_mtab_file = "/proc/mounts";
		}
	}
	do {
		const char *device;
		const char *mount_point;

		if (mount_table) {
			if (!(mount_entry = getmntent(mount_table))) {
				endmntent(mount_table);
				break;
			}
		} else {
			if (!(mount_entry = find_mount_point(mount_point, bb_path_mtab_file))) {
				bb_error_msg("%s: can't find mount point.", mount_point);
			SET_ERROR:
				continue;
			}
		}

		device = mount_entry->mnt_fsname;
		mount_point = mount_entry->mnt_dir;
	
		if (statfs(mount_point, &s) != 0) {
			bb_error_msg("%s: %s", mount_point, strerror(errno));
			goto SET_ERROR;
		}

		if ((s.f_blocks > 0) || !mount_table){
			blocks_used = s.f_blocks - s.f_bfree;
			blocks_percent_used = 0;
			if (blocks_used + s.f_bavail) {
				blocks_percent_used = (((long long) blocks_used) * 100
									+ (blocks_used + s.f_bavail)/2
									) / (blocks_used + s.f_bavail);
			}

			if (strcmp(device, "rootfs") == 0) {
				continue;
			} else if (strcmp(device, "/dev/root") == 0) {
				/* Adjusts device to be the real root device,
				* or leaves device alone if it can't find it */
				if ((device = find_block_device("/")) == NULL) {
					goto SET_ERROR;
				}
			}

			row[0] = device;
			size   = xmake_message("%s", make_human_readable_str(s.f_blocks, s.f_bsize, 0 ));
			row[1] = STR_YES;	
			row[2] = size; 
			used   = xmake_message("%s", make_human_readable_str( (s.f_blocks - s.f_bfree), s.f_bsize, 0));
			row[3] = used;
			avail  = xmake_message("%s", make_human_readable_str(s.f_bavail, s.f_bsize, 0));
			row[4] = avail;
			snprintf(percent, 5, "%3ld%%", blocks_percent_used); 
			row[5] = percent;
			row[6] = mount_entry->mnt_type;
			row[7] = mount_point;
			row[8] = mount_entry->mnt_opts;

			
			Fdata->device = row[0];
			Fdata->ptr_found = &found;
			
			/* Ricerca se presente nella lista */
			if (strcmp(bb_path_mtab_file, "/etc/fstab") == 0) {
				gtk_tree_model_foreach (GTK_TREE_MODEL( gtk_tree_view_get_model(GTK_TREE_VIEW(clist))), 
									 (GtkTreeModelForeachFunc)tot_list_foreach_func, Fdata); 
				row[1] = STR_NO;	
				row[2] = "";	
				row[3] = "";
				row[4] = "";
				row[5] = "";
	
				
			}

			if (found == 0) {
				add_data_to_list(clist,(char **) row, 9, NO_AUTOSCROLL);
			}
			found = 0;

			if(strcmp(device, mount_entry->mnt_fsname) != 0) {
				xfree(&device);
			}
			xfree(&size);
			xfree(&used);
			xfree(&avail);
		}
	
	} while (1);
	
	if (strcmp(bb_path_mtab_file, "/etc/fstab") != 0) {
		bb_path_mtab_file = "/etc/fstab";
		goto SECOND_PASS;
	}
	xfree(&Fdata);
}

void do_umount(GtkWidget *widget, GtkWidget *list)
{
	DBG_FUNCTION_NAME;

	assert(widget != NULL);
	if (strcmp(is_mounted, STR_YES) != 0) {
		return;
	}
	if(umount(mount_path) == -1)
		dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1, STR_CANNOT_UNMOUNT_S_S, mount_path, strerror(errno));
	else
		erase_mtab(mount_path);
	clear_list(list);
	df_list(list);
}

/*
 * Mini mount implementation for busybox
 *
 * Copyright (C) 1995, 1996 by Bruce Perens <bruce@pixar.com>.
 * Copyright (C) 1999-2004 by Erik Andersen <andersen@codepoet.org>
 * Copyright (C) 2005 by Rob Landley <rob@landley.net>
 *
 * Licensed under GPLv2 or later, see file LICENSE in this tarball for details.
 */

/* Consume standard mount options (from -o options or --options).
 * Set appropriate flags and collect unrecognized ones as a comma separated
 * string to pass to kernel */ 

struct {
	const char *name;
	long flags;
} static const  mount_options[] = {
	{"loop", 0},
	{"defaults", 0},
	{"noauto", 0},
	{"ro", MS_RDONLY},
	{"rw", ~MS_RDONLY},
	{"nosuid", MS_NOSUID},
	{"suid", ~MS_NOSUID},
	{"dev", ~MS_NODEV},
	{"nodev", MS_NODEV},
	{"exec", ~MS_NOEXEC},
	{"noexec", MS_NOEXEC},
	{"sync", MS_SYNCHRONOUS},
	{"async", ~MS_SYNCHRONOUS},
	{"remount", MS_REMOUNT},
	{"atime", MS_NOATIME},
	{"noatime", MS_NOATIME},
	{"diratime", MS_NODIRATIME},
	{"nodiratime", MS_NODIRATIME},
	{"bind", MS_BIND},
	{"move", MS_MOVE}
};

/* Uses the mount_options list above */
static void parse_mount_options(char *options, int *flags, char **strflags)
{
	/* Loop through options */
	for(;;) {
		int i;
		char *comma = strchr(options, ',');

		if(comma) *comma = 0;

		/* Find this option in mount_options */
		for(i = 0; i < (sizeof(mount_options) / sizeof(*mount_options)); i++) {
			if(!strcasecmp(mount_options[i].name, options)) {
				long fl = mount_options[i].flags;
				if(fl < 0) *flags &= fl;
				else *flags |= fl;
				break;
			}
		}
		/* Unrecognized mount option? */
		if(i == (sizeof(mount_options) / sizeof(*mount_options))) {
			/* Add it to strflags, to pass on to kernel */
			i = *strflags ? strlen(*strflags) : 0;
			*strflags = xrealloc(*strflags, i+strlen(options)+2);
			/* Comma separated if it's not the first one */
			if(i) (*strflags)[i++] = ',';
			strcpy((*strflags)+i, options);
		}
		/* Advance to next option, or finish */
		if(comma) {
			*comma = ',';
			options = ++comma;
		} else break;
	}
}

void do_mount(GtkWidget *widget, GtkWidget *list)
{
	char *string_flags = 0;
	char *fsType = 0;
	char *buf = 0;
	char *files[] = {"/etc/filesystems", "/proc/filesystems", 0};
	int i = 0, rc = 1;
	int flags=0xc0ed0000;	// Needed for linux 2.2, ignored by 2.4 and 2.6.
	FILE *file = 0,*f = 0;
	char *bb_path_mtab_file	= "/etc/mtab";
	struct mntent m;

	DBG_FUNCTION_NAME;

	assert(widget != NULL);
	
	if (strcmp(is_mounted, STR_NO) != 0) {
		return;
	}

	parse_mount_options(mount_opts, &flags, &string_flags);

	if ( mount_type[0] != '\0' || strcmp(mount_type, "auto") != 0 ) {
		fsType = mount_type;
		goto mount_it_now;
	}

	/* Loop through filesystem types until mount succeeds or we run out */

	for(i = 0; files[i] && rc; i++) {
		f = fopen(files[i], "r");
		if(!f) continue;
		/* Get next block device backed filesystem */
		for(buf = 0; (buf = fsType = get_chunk_from_file(f, NULL)); free(buf))
		{
			/* Skip funky entries in /proc */
			if(!strncmp(buf,"nodev",5) && isspace(buf[5])) continue;
			
			while(isspace(*fsType)) fsType++;
			if(*buf=='#' || *buf=='*') continue;
			if(!*fsType) continue;
mount_it_now:
			/* Okay, try to mount */

			for(;;) {
				rc = mount(mount_device, mount_path, fsType, flags, string_flags);
				if(!rc || (flags&MS_RDONLY) || (errno!=EACCES && errno!=EROFS))
					break;
				dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1,
								STR_S_IS_WRITE_PROTECTED_MOUNTING_RO, mount_device);
				flags|=MS_RDONLY;
			}
			if(!rc || !f) break;
		}
		if(f) fclose(f);
		if(!f || !rc) break;
	}

	/* If the mount was successful, and we're maintaining an old-style
		* mtab file by hand, add new entry to it now. */
	if (!rc) {
		FILE *mountTable = setmntent(bb_path_mtab_file, "a+");

		if(!mountTable) {
			dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1,"%s:\n%s",bb_path_mtab_file, strerror(errno));
		} else {
			/* Remove trailing / (if any) from directory we mounted on */
			int length=strlen(mount_path);
			if(length>1 && mount_path[length-1] == '/')
				mount_path[length-1]=0;

			/* Fill out structure (should be ok to re-use existing one)*/
			m.mnt_fsname=mount_device;
			m.mnt_dir=mount_path;
			m.mnt_type=fsType ? fsType : "--bind";
			for( i = 0; i < xstrlen(mount_opts); i++) {
				if(mount_opts[i] == 'r' && (!mount_opts[i+2] || mount_opts[i+2] == ',')) {
					mount_opts[i+1] = (flags & MS_RDONLY) ? 'o' : 'w';
				}
			}
			m.mnt_opts = mount_opts;
			m.mnt_freq = 0;
			m.mnt_passno = 0;

			/* Write and close */
			addmntent(mountTable, &m);
			endmntent(mountTable);
			/*m.mnt_opts = xfree(m.mnt_opts);*/
		}
	} else {
		/* Mount failed.  Clean up */
		if (errno == EPERM)
			dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1,"%s", strerror(errno));
	}
	/* We couldn't free this earlier becase fsType could be in buf */
	xfree(&buf);
	if(file) endmntent(file);
	if(rc) {
		dialog_window(STR_ERROR, ONE_BUTTON, NULL, 1,
						STR_MOUNTING_S_ON_S_FAILED, mount_device, mount_path); 
	}
	clear_list(list);
	df_list(list);
}

void df_f(GtkWidget *tree)
{
	GtkTreeIter iter;
	GtkTreeModel *model;
	GtkTreeSelection *selection;

	if (GTK_IS_TREE_VIEW(tree))
		selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
	else
		selection = GTK_TREE_SELECTION(tree);

	if (gtk_tree_selection_get_selected ( selection, &model, &iter)) {
		gtk_tree_model_get (model, &iter, 0, &mount_device, 2, &is_mounted, 6, &mount_type,
							7, &mount_path, 8, &mount_opts, -1);
	
		debug_err_msg("%s(): you selected : %s %s %s %s %s",__FUNCTION__,
						mount_device, mount_path, mount_type, is_mounted, mount_opts);
	}
}

void df_window(void)
{
	GtkWidget *df_win;
	GtkWidget *clist;
	GtkWidget *table;
	GtkWidget *bbox;
	GtkWidget *button;

	df_win = window_create(640, 240, 2, STR_FS_INFO, GTK_SIGNAL_FUNC(gtk_main_quit), NULL,  MODAL);
	set_position (GTK_WINDOW (df_win), GTK_WIN_POS_CENTER);
	table = table_create(4, 4,  df_win, 2, HOMOGENEUS);
	clist = string_list_create_on_table(9, NULL, table, 0, 4, 0, 3, SHOW, NOT_EDITABLE,
										" "STR_FILESYSTEM, 0.0, " "STR_MOUNTED, 0.0, " "STR_SIZE, 1.0, " "STR_USED, 1.0,
										" "STR_AVAILABLE, 1.0, " %", 1.0, " "STR_TYPE, 0.0, " "STR_MOUNTPOINT, 0.0,
										" "STR_MOUNT_OPTIONS, 0.0);
	signal_connect (GTK_TREE_SELECTION(gtk_tree_view_get_selection (GTK_TREE_VIEW (clist))), "changed",
						G_CALLBACK (df_f), NULL );
	signal_connect (GTK_TREE_VIEW (clist), "row-activated", G_CALLBACK (df_f), NULL);
	bbox = default_button_box_create_on_table( HORIZONTAL, table, 0, 4, 3, 4);
	button = button_create_in_container(STR_MOUNT, GTK_SIGNAL_FUNC (do_mount), clist, bbox);
	button = button_create_in_container(STR_UNMOUNT, GTK_SIGNAL_FUNC(do_umount), clist, bbox);
	button = button_create_in_container(STR_QUIT, GTK_SIGNAL_FUNC(window_close) , df_win, bbox);	
	gtk_widget_show_all(df_win);
	df_list(clist);
	main_loop(&df_win);
	return;
}
