/*
*
* Gtk+2.x  Utility routines: Text Editor Window
* Copyright (C) 2002-2006 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Please read the COPYING and README file!!!
*
* Report bugs to <tito-wolit@tiscali.it>
*
*/

#include <gtk/gtk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <strings.h>

static int edit_lock = 0;
static GtkWidget *edit_window;
static GtkWidget *edit_table;	
static GtkWidget *edit_text;
static struct stat sourcefileinfo;
static int text_modified = 0;
static char *path = NULL;

/********************* Text file edit Widget *******************************/

void edit_wrap_text (void)
{
	gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(edit_text),
								(gtk_text_view_get_wrap_mode (GTK_TEXT_VIEW(edit_text)) != GTK_WRAP_NONE) ?
								GTK_WRAP_NONE :
								GTK_WRAP_WORD);
}

static void save_file(char *filename, GtkWidget *text)
{
	/*   Inspired by files.c of nano editor 1.2.2                            *
	*                                                                        *
	*   Copyright (C) 1999-2003 Chris Allegretta GPL                         *
	*   realeased under the terms of the GPL                                 */

	GtkTextBuffer	*buffer;
	GtkTextIter		start;
	GtkTextIter		end;
	char *data = NULL;
	int flag;
	int fd;
	mode_t mode = 0600;
	FILE *file = NULL;
	struct stat fileinfo;

	if (filename[0] == '\0' || filename == NULL ) {
		dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_NO_FILE_NAME);
		return;
	}

	if (stat(filename, &fileinfo) == 0) {
		/* File exists */
		if (S_ISDIR(fileinfo.st_mode) || S_ISCHR(fileinfo.st_mode) || S_ISBLK(fileinfo.st_mode)) {
			if(S_ISDIR(fileinfo.st_mode)) {
				dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_S_IS_A_DIR, filename);
			} else {
				dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_FILE_S_IS_A_DEVICE_FILE, filename);
			}
			return;
		}
		mode = fileinfo.st_mode;
		flag = O_WRONLY| O_TRUNC ;
	} else {
		flag = O_WRONLY| O_CREAT;
	}

	if ((fd = open(filename, flag, mode)) == -1) {
		dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, "%s: %s", filename, strerror(errno));
		return;
	}
	if ((file = fdopen(fd, "wb"))) {
		buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (text));
		gtk_text_buffer_get_bounds (buffer, &start, &end);
		data = g_locale_from_utf8(gtk_text_buffer_get_text(buffer, &start, &end,
									/*include_hidden_chars*/ FALSE), -1, NULL, NULL, NULL);
		if (data) {
			fprintf(file,"%s", data);
			xfree(&data);
		} else {
			dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_CANNOT_SAVE_S_UNKNOWN_ERROR, filename);
		}
		fclose(file);
	} else {
		dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_ERROR_WRITING_FILE_S_S, filename, strerror(errno));
	}
	close(fd);
}

static void text_changed_callback(void)
{
	text_modified = 1;
}

static int overwrite(char *msg_str)
{
	DBG_FUNCTION_NAME;

	if (dialog_window(STR_WARNING, TWO_BUTTONS, NULL, 2, "%s", msg_str) == 1) {
		return 1;
	}
	return 0;
}

static void save_common(void)
{
	mode_t mask;
	struct stat destfileinfo;
		
	/* File exists */
	/* g_locale_from_utf8(path,-1, NULL, NULL, NULL);*/
	if (stat(path, &destfileinfo) == 0) {
		if (overwrite(STR_DEST_FILE_EXIST) == 1) {
			return;
		}
		mask = destfileinfo.st_mode;
		/* File is newer */
		if (sourcefileinfo.st_mtime!=0 && sourcefileinfo.st_mtime > destfileinfo.st_mtime ) {
			if (overwrite(STR_DEST_FILE_NEWER) == 1)
				return;
		}
	} else {
		if (sourcefileinfo.st_mode != 0) {
			mask = sourcefileinfo.st_mode;
		} else {
			mask = 0600; /* Use default file permissions */
		}
	}
	/* write file*/ 
	save_file(path, edit_text);
	text_modified = 0;
	/* Set permission */
	if (chmod(path, mask) == -1) {
		dialog_window_and_beep(STR_ERROR, ONE_BUTTON, NULL, 1, STR_CANNOT_SET_FILE_PERMISSIONS_ON_S_S, path, strerror(errno));
	}
	/* Set ownership is not needed as it is set to that of the current process */
	gtk_window_set_title(GTK_WINDOW(edit_window), path);
}

static void save_as_cb(void)
{
	char *new_path;
	
	if ((new_path = default_file_selector( STR_SAVE_AS":", GTK_FILE_CHOOSER_ACTION_SAVE)) == NULL) {
		return;
	}
	xfree(&path);
	path = new_path;
	save_common();
}

static void save_cb(void)
{
	if (text_modified == 0) {
		return;
	}
	if (path == NULL) {
		/* No file name: ask for it */ 
		if ((path = default_file_selector( STR_SAVE_AS":", GTK_FILE_CHOOSER_ACTION_SAVE)) == NULL) {
			/* Cancel button */
			return;
		}
	}
	save_common();
}

static void open_common(void)
{
	ProgressData * pdata = NULL;
	FILE *file = NULL;
	GtkTextIter		iter;
	char *title = NULL;

	DBG_FUNCTION_NAME;

	/* During file reading delete_event and quit signals are ignored */
	edit_lock = 1;
	gtk_text_buffer_get_iter_at_offset (gtk_text_view_get_buffer (GTK_TEXT_VIEW (edit_text)), &iter, 0);

	if (path && g_utf8_validate(path , -1, NULL) != TRUE) {
		title = g_locale_to_utf8(path, -1, NULL, NULL, NULL);
	} else {
		title = xstrdup(path);
	}

	if (path && stat(path, &sourcefileinfo) == 0) {
		if ((file = open_file(path)) != NULL) {
			/* Errors in opening the file are showed by open_file itself */
			gtk_window_set_title(GTK_WINDOW(edit_window), title);
			pdata = progressbar_create_on_table(PULSE, edit_table, 0, 10, 20, 21 );
			gtk_widget_show(pdata->pbar);
			read_from_stream_and_display_text(edit_text, file, sourcefileinfo.st_size, &pdata->value);
			xfclose_nostdin(file);
			progress_destroy(pdata);
		} else {
			/* Some error occurred in opening the file: 
			S_ISDIR, S_ISCHR, S_ISBLK, so create create an untitled empty edit window */
			gtk_window_set_title(GTK_WINDOW(edit_window), STR_UNTITLED );
		}
	} else {
		/*create an empty edit window titled or untitled */
		gtk_window_set_title(GTK_WINDOW(edit_window),(path) ? title : STR_UNTITLED);
	}
	xfree(&title);
	text_modified = 0;
	edit_lock = 0;
}

static void new_cb(void)
{
	if (!edit_lock) {
		if (text_modified){
			if (!path) {
				/* No file name: ask for it */ 
				if ((path = default_file_selector( STR_SAVE_AS":", GTK_FILE_CHOOSER_ACTION_SAVE)) ==NULL) {
					/* Cancel button */
					return;
				}
			}	
			save_common();
		}
		gtk_window_set_title(GTK_WINDOW(edit_window), STR_UNTITLED);
		clear_the_buffer(edit_text, &text_modified);
		xfree(&path);
		text_modified = 0;
		sourcefileinfo.st_mtime = 0;
		sourcefileinfo.st_mode = 0;
	}
}

static void open_cb(void)
{
	if (!edit_lock) {
		new_cb();
		if ((path = default_file_selector( STR_OPEN":", GTK_FILE_CHOOSER_ACTION_OPEN))) {
			open_common();
		}
	}
}

static void quit_cb(void)
{	
	DBG_FUNCTION_NAME;

	if (!edit_lock) {
		if (text_modified && dialog_window(STR_SAVE_FILE, TWO_BUTTONS, NULL, 2, STR_FILE_MODIFIED_WANT_TO_SAVE) == 0) {
			if (!path) {
				/* No file name: ask for it */ 
				if ((path = default_file_selector( STR_SAVE_AS":", GTK_FILE_CHOOSER_ACTION_SAVE)) == NULL) {
					/* Cancel button */
					return;
				}
			}	
			save_common();
		}
		gtk_main_quit();
		gtk_widget_destroy(edit_window);
	}
}

static GtkActionEntry text_edit_entries[] = {
	/* name,		stock id,	label,				accelerator,		tooltip,	callback */
	{ "FileMenu",	NULL,		MNU_STR_FILE,		NULL,				NULL,		NULL},
	{ "OptionsMenu",NULL,		MNU_STR_OPTIONS,	NULL,				NULL,		NULL},	
	{ "HelpMenu",	NULL,		MNU_STR_HELP,		NULL,				NULL,		NULL},
	{ "New",		NULL,		MNU_STR_NEW, 		SHC_FILE_NEW,		NULL,		G_CALLBACK (new_cb) },   
	{ "Open",		NULL,		MNU_STR_OPEN,		SHC_FILE_OPEN,		NULL,		G_CALLBACK (open_cb) }, 
	{ "Save",		NULL,		MNU_STR_SAVE,		SHC_FILE_SAVE,		NULL,		G_CALLBACK (save_cb) },
	{ "SaveAs",		NULL,		MNU_STR_SAVE__AS,	SHC_FILE_SAVE_AS,	NULL,		G_CALLBACK (save_as_cb) }, 
	{ "Quit",		NULL,		MNU_STR_QUIT,		SHC_FILE_QUIT,		NULL,		G_CALLBACK (quit_cb) },
	{ "About",		NULL,		MNU_STR_ABOUT,		SHC_HELP_ABOUT,		NULL,		G_CALLBACK (about_cb) },
};

static GtkToggleActionEntry text_edit_check_entries[] = {
{	
	"Wrap", NULL,						/* name, stock id */
	MNU_STR_WRAP, SHC_OPTIONS_WRAP,			/* label, accelerator */     
	NULL,								/* tooltip */
	G_CALLBACK (edit_wrap_text),		/* callback */ 
	FALSE },							/* is_active */
};

static const gchar *text_edit_ui_info = 
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='New'/>"
"      <menuitem action='Open'/>"
"      <menuitem action='Save'/>"
"      <menuitem action='SaveAs'/>"
"      <separator/>"
"      <menuitem action='Quit'/>"
"    </menu>"
"    <menu action='OptionsMenu'>"
"      <menuitem action='Wrap'/>"
"    </menu>"
"    <menu action='HelpMenu'>"
"      <menuitem action='About'/>"
"    </menu>"
"  </menubar>"
"</ui>";

static struct MenuEntry text_edit_menu_entries[] = {
{ text_edit_entries,	NULL,						NULL,			G_N_ELEMENTS (text_edit_entries),			0, NULL},
{ NULL,					text_edit_check_entries,	NULL,			G_N_ELEMENTS (text_edit_check_entries),		0, NULL},
{ NULL,					NULL,						NULL,			0,											0, NULL},
/* Set last to NULL */ 
};

void textedit_window_create(char * title, int x, int y, int border, char *filename)
{

	GtkWidget *menubar;

	DBG_FUNCTION_NAME;
	path = xstrdup(filename);
	edit_window = window_create( x, y, border, title,  GTK_SIGNAL_FUNC (quit_cb), NULL,  MODAL);
	edit_table = table_create( 20, 21,GTK_WIDGET(edit_window), 0, HOMOGENEUS);
	menubar = main_menu_create(GTK_WIDGET (edit_window), text_edit_menu_entries, text_edit_ui_info, NULL);
	gtk_table_attach_defaults (GTK_TABLE(edit_table),GTK_WIDGET(menubar), 0, 20, 0, 1);

	edit_text = textpad_create_on_table(EDITABLE, GTK_SIGNAL_FUNC(text_changed_callback),
									&text_modified, NO_WRAP, SCROLL, edit_table,  0, 20, 1, 20);
	gtk_widget_show_all(GTK_WIDGET(edit_window));
	open_common();
	main_loop(&edit_window);
	xfree(&path);
}

/***************End of Text file edit Widget ********************************/
