/*
*
* Gtk+1.2-Gtk+2.2 Utility routines
* Copyright (C) 2002-2003 by Tito Ragusa <tito-wolit@tiscali.it>
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Please read the COPYING and README file!!!
*
* Report bugs to <tito-wolit@tiscali.it>
*
*/

#include <gtk/gtk.h>

/* This are defines in mygtk.h:
	signal_connect(GTK_OBJECT (button), "signal", GTK_SIGNAL_FUNC(func), func_data);
	set_editable(x,TRUE/FALSE);
	get_radio_button_group(GTK_RADIO_BUTTON (x));
	set_position (GTK_WINDOW (x), GTK_WIN_POS_CENTER);
	clear_list(x);
*/

#ifdef L_window_close
void window_close(GtkWidget *widget, GtkWidget *window)
{
	DBG_FUNCTION_NAME;

	assert(widget!=NULL);
	assert(window!=NULL);
	
	gtk_main_quit();
#ifdef USE_GTK_2_0	
	if(widget!=window) {
		gtk_widget_destroy(window);
	}
#endif
	return;
}
#endif

#ifdef L_main_loop
GtkWidget *main_loop(GtkWidget *window)
{
	if (!GTK_WIDGET_VISIBLE (window))
    	gtk_widget_show_all (window);
	gtk_main();
#ifndef USE_GTK_2_0
	if(GTK_IS_WIDGET (window))
		gtk_widget_destroy(window);
#endif
	return window=NULL;
}
#endif

#ifdef L_attach_to_table
void attach_to_table(GtkWidget *table, GtkWidget *widget, int start_col, int end_col, int start_row, int end_row)
{
	debug(if (start_col <  0)err_msg("%s() start_col <0",__FUNCTION__));
	debug(if (end_col   <= 0)err_msg("%s() end_col  <=0",__FUNCTION__));
	debug(if (start_row <  0)err_msg("%s() start_row <0",__FUNCTION__));
	debug(if (end_col   <= 0)err_msg("%s() end_row  <=0",__FUNCTION__));
	debug(if (end_col <= start_col )err_msg("%s() end_col <= start_col",__FUNCTION__));
	debug(if (end_row <= start_row )err_msg("%s() end_row <= start_row",__FUNCTION__));

	gtk_table_attach_defaults (GTK_TABLE(table), GTK_WIDGET(widget), start_col, end_col, start_row, end_row);
}
#endif

/********************* Text box widget ***************************************/

#ifdef L_text_box_create_on_table
GtkWidget *text_box_create_on_table(char *text, int editable, GtkWidget *table,
								 int start_col, int end_col, int start_row, int end_row)
{
	GtkWidget *entry_box;

	entry_box = gtk_entry_new();
	attach_to_table (table, entry_box, start_col, end_col, start_row, end_row);
	if(text!=NULL)
		gtk_entry_set_text(GTK_ENTRY(entry_box),text);
	set_editable(entry_box,(editable==EDITABLE)?TRUE:FALSE);
	return entry_box;
}
#endif
/********************* End of Text box widget ********************************/

/********************* Standard radio button widget **************************/

#ifdef L_radio_create
GtkWidget *radio_create( GtkWidget *last_of_group , char *label,
			 GtkSignalFunc func, gpointer func_data, int active )
{
	GtkWidget *button;

	if( last_of_group==NULL)
		button = gtk_radio_button_new_with_label(NULL, label);
	else
		button = gtk_radio_button_new_with_label(get_radio_button_group(GTK_RADIO_BUTTON (last_of_group)), label);
	
	if(func!=NULL)
		signal_connect (GTK_OBJECT (button), "toggled", GTK_SIGNAL_FUNC(func), func_data);

	debug(if (func==NULL) err_msg("%s() GtkSignalFunc func == NULL",__FUNCTION__));

	if(active==ACTIVE)
	{
		/* gtk_toggle_button_set_active is not needed on first button (last_of_group==NULL)
			because it is set active by default, but to be sure.......
		*/
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),TRUE);
		gtk_toggle_button_toggled (GTK_TOGGLE_BUTTON (button));
	}
	return button;
}
#endif
/****************End of Standard radio button widget *************************/

/********************* Radio button on table widget **************************/
#ifdef L_radio_create_on_table
GtkWidget *radio_create_on_table(GtkWidget *last_of_group, char *label, GtkSignalFunc func, gpointer func_data,
										int active, GtkWidget *table, int start_col, int end_col,
										int start_row, int end_row )
{
	GtkWidget *button;
	button = radio_create(last_of_group, label, func, func_data , active );
	attach_to_table (table, button, start_col, end_col, start_row, end_row);
	return button;
}
#endif

#ifdef L_default_radio_create_on_table
GtkWidget *default_radio_create_on_table(GtkWidget *last_of_group, char *label, GtkSignalFunc func,
										int active, GtkWidget *table, int start_col, int end_col,
										int start_row, int end_row )
{
	GtkWidget *button;
	button = radio_create(last_of_group, label, func, label ,active );
	attach_to_table (table, button, start_col, end_col, start_row, end_row);
	return button;
}
#endif
/****************End of radio button on table widget *************************/

/********************* Check button widgets **************************/
#ifdef L_check_create
GtkWidget *check_create(char *label, GtkSignalFunc func, gpointer func_data, int active )
{
	GtkWidget *button;

	button = gtk_check_button_new_with_label(label);
	
	if(func!=NULL)
		signal_connect (GTK_OBJECT (button), "toggled", GTK_SIGNAL_FUNC(func), func_data);
	
	debug(if(func==NULL) err_msg("%s(): GtkSignalFunc func == NULL",__FUNCTION__));
	
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button),(active==ACTIVE)?TRUE:FALSE);

	return button;
}
#endif

#ifdef L_check_create_on_table
GtkWidget *check_create_on_table(char *label, GtkSignalFunc func, gpointer func_data, int active,
			 GtkWidget *table, int start_col, int end_col, int start_row, int end_row )
{
	GtkWidget *button;
	button = check_create(label, func, func_data ,active );
	attach_to_table (table, button, start_col, end_col, start_row, end_row);
	return button;
}
#endif

#ifdef L_default_check_create_on_table
GtkWidget *default_check_create_on_table(char *label, GtkSignalFunc func, int active,
			 GtkWidget *table, int start_col, int end_col, int start_row, int end_row )
{
	GtkWidget *button;
	button = check_create(label, func, label ,active );
	attach_to_table (table, button, start_col, end_col, start_row, end_row);
	return button;
}
#endif
/***************End of Check button widgets **********************************/

/********************* Standard button widget ********************************/
#ifdef L_button_create
GtkWidget *button_create(char *label, GtkSignalFunc func, gpointer func_data )
{
	GtkWidget *button;

	button = gtk_button_new_with_label (label);
	if(func!=NULL)
		signal_connect (GTK_OBJECT (button), "clicked", GTK_SIGNAL_FUNC (func), func_data);
	
	debug(if(func==NULL) err_msg("%s() GtkSignalFunc func == NULL",__FUNCTION__));

	return button;
}
#endif
/****************End of Standard button widget *******************************/

/********************* Button on table widget ********************************/
#ifdef L_button_create_on_table
GtkWidget *button_create_on_table(char *label, GtkSignalFunc func, gpointer func_data,
			 GtkWidget *table, int start_col, int end_col, int start_row, int end_row )
{
	GtkWidget *button;

	button = button_create(label, func, func_data );
	attach_to_table (table, button, start_col, end_col, start_row, end_row);
	return button;
}
#endif
/****************End of button on table widget ******************************/

/********************* Button in container widget ***************************/
#ifdef L_button_create_in_container
GtkWidget *button_create_in_container(char *label, GtkSignalFunc func, gpointer func_data, GtkWidget *container)
{
	GtkWidget *button;

	button = button_create(label, func, func_data);
	gtk_container_add (GTK_CONTAINER (container), button);
	return button;
}
#endif
/*************End of Button in container widget *****************************/

/********************* Buttonbox on table widget ****************************/
#ifdef L_button_box_create_on_table
GtkWidget *button_box_create_on_table(	int orientation, int layout,  int border,
										int spacing, int child_size_x, int child_size_y, GtkWidget *table,
										int start_col, int end_col, int start_row, int end_row )
{
	GtkWidget *bbox;

	debug(if(spacing<=0) err_msg("%s() spacing <= 0",__FUNCTION__));
	debug(if(border<=0) err_msg("%s() border <= 0",__FUNCTION__));
	debug(if(child_size_x<=0) err_msg("%s() child_size_x <= 0",__FUNCTION__));
	debug(if(child_size_y<=0) err_msg("%s() child_size_y <= 0",__FUNCTION__));
	
	if(orientation==HORIZONTAL)
		bbox = gtk_hbutton_box_new ();
	else if (orientation==VERTICAL)
		bbox = gtk_vbutton_box_new ();
	else
	{
		/* Bad Value */	
		debug(err_msg("%s(): orientation != HORIZONTAL or VERTICAL",__FUNCTION__));
		return NULL;
	}

	gtk_container_set_border_width (GTK_CONTAINER (bbox), border);
	attach_to_table (table, bbox, start_col, end_col, start_row, end_row);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (bbox), layout);
	
	#ifdef USE_GTK_2_0
	gtk_box_set_spacing (GTK_BOX (bbox), spacing);
	
	/* To silence a compiler warning about unused vars */
	child_size_x=child_size_y=0;
	#else
	gtk_button_box_set_spacing (GTK_BUTTON_BOX (bbox), spacing);
	gtk_button_box_set_child_size (GTK_BUTTON_BOX (bbox), child_size_x, child_size_y);
	#endif
	
	return bbox;
}
#endif

#ifdef L_default_button_box_create_on_table
GtkWidget *default_button_box_create_on_table(	int orientation, GtkWidget *table,
										 		int start_col, int end_col,
												int start_row, int end_row )
{
	GtkWidget *bbox;

	/* These default values look good to me. If you don't like it
		use button_box_create_on_table()	to set your own */
	bbox = button_box_create_on_table( orientation,  GTK_BUTTONBOX_END, 10, 20, 85, 30,
								 		table, start_col, end_col, start_row, end_row );
	return bbox;
}
#endif
/***************End of Buttonbox on table widget ****************************/

/********************* Label on table widget ********************************/
#ifdef L_label_create_on_table
GtkWidget *label_create_on_table(char *label_text,GtkWidget *table,
				int start_col, int end_col, int start_row, int end_row )
{
	GtkWidget *label;
	label=gtk_label_new(label_text);
	attach_to_table (table, label, start_col, end_col, start_row, end_row);
	return label;
}
#endif
/***************End of Label on table widget ********************************/

/********************* Standard table widget ********************************/
#ifdef L_table_create
GtkWidget *table_create(int cols , int rows,  GtkWidget* window , int border, int homogeneus)
{
	GtkWidget *table;

	debug(if(rows <= 0) err_msg("%s() rows <=0",__FUNCTION__));
	debug(if(cols <= 0) err_msg("%s() cols <=0",__FUNCTION__));
	debug(if(border < 0) err_msg("%s() border < 0",__FUNCTION__));
	debug(if(window==NULL)err_msg("%s() GtkWidget *window == NULL",__FUNCTION__));

	table = gtk_table_new (rows, cols, TRUE);

	#ifdef USE_GTK_2_0
	/* THIS MAKES THINGS LOOK BETTER */
	gtk_table_set_homogeneous(GTK_TABLE(table), homogeneus);
	#else
	/*Ignored: to shut up a compiler warning */
	homogeneus=IGNORED;
	#endif

	gtk_container_add (GTK_CONTAINER (window), table);
	gtk_container_set_border_width (GTK_CONTAINER (table), border);
	return table;
}
#endif

/********************* Standard table widget ********************************/

/********************* Standard window widget ********************************/
#ifdef L_window_create
GtkWidget *window_create( int x, int y, int border, const char *title, GtkSignalFunc func,  void *func_data, int modal)
{
		GtkWidget	*window;
		
		debug(if(x<=0) err_msg("%s() winsize x <=0",__FUNCTION__));
		debug(if(y<=0) err_msg("%s() winsize y <=0",__FUNCTION__));
		debug(if(border < 0) err_msg("%s() border < 0",__FUNCTION__));
		debug(if(func==NULL)err_msg("%s() GtkSignalFunc func == NULL using gtk_true as callback",__FUNCTION__));
		
		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	
		if(func!=NULL)
		{
			signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC (func), func_data);
			/*signal_connect(GTK_OBJECT(window), "destroy",      GTK_SIGNAL_FUNC (func), func_data);*/
		} else {
			signal_connect(GTK_OBJECT(window), "delete_event", GTK_SIGNAL_FUNC (gtk_true), func_data);
			/*signal_connect(GTK_OBJECT(window), "destroy",      GTK_SIGNAL_FUNC (gtk_true), func_data);*/
		}
		#ifdef USE_GTK_2_0
		gtk_window_set_default_size(GTK_WINDOW(window), x, y);
		#else
		gtk_widget_set_usize(GTK_WIDGET(window), x, y);
		#endif
		
		if(title!=NULL)
			gtk_window_set_title(GTK_WINDOW(window), title);
		else
			gtk_window_set_title(GTK_WINDOW(window), "");
		gtk_container_set_border_width (GTK_CONTAINER (window), border);
		gtk_window_set_modal(GTK_WINDOW(window),(modal==MODAL) ? TRUE : FALSE);
		return window;
}
#endif

/********************* Standard window widget ********************************/

/********************* Frame on table widget ********************************/
#ifdef L_frame_create_on_table
GtkWidget *frame_create_on_table(char *label, GtkWidget *table,
											 int start_col,int end_col, int start_row, int end_row )
{
	GtkWidget *frame;

	if(label==NULL) label="";
	frame = gtk_frame_new(label);
	attach_to_table (table, frame, start_col, end_col, start_row, end_row);
	return frame;
}
#endif
/**************End of Frame on table widget ********************************/

/************** Create a separator *****************************************/
#ifdef L_separator_create_on_table
GtkWidget *separator_create_on_table(int orientation, GtkWidget *table, int start_col,int end_col, int start_row, int end_row )
{
	GtkWidget *separator;

	if(orientation==HORIZONTAL)
		separator=gtk_hseparator_new();
	else if(orientation==VERTICAL)
		separator=gtk_vseparator_new();
	else
	{
		debug(err_msg("%s(): orientation != HORIZONTAL or VERTICAL",__FUNCTION__));
		return NULL;
	}
	
	attach_to_table (table, separator, start_col, end_col, start_row, end_row);
	return separator;
}
#endif
/************** End of Create a separator **********************************/

/************** Create a Multi column scrolled list ************************/
#ifdef L_add_data_to_list
#ifdef USE_GTK_2_0
void add_data_to_list(GtkWidget * tree, char *data_string[], int n_columns)
{
	GtkTreeIter			iter;
	GtkListStore		*store;
	char * data=NULL;
	int i;
	
	DBG_FUNCTION_NAME;
	
	store = (GtkListStore *)gtk_tree_view_get_model (GTK_TREE_VIEW(tree));
	gtk_list_store_append (GTK_LIST_STORE(store), &iter);

	for(i=0 ; i < n_columns; i++, *data_string++)
	{
		if((data = g_locale_to_utf8(*data_string, /*xstrlen(*data_string)*/-1, NULL, NULL, NULL))!=NULL)
		{
			if(GTK_IS_WIDGET(tree))
				gtk_list_store_set (GTK_LIST_STORE(store), &iter, i, data, -1);
			data=xfree(data);
		}
	}
}
#else
void add_data_to_list(GtkWidget * tree, char *data_string[], int n_columns)
{
	DBG_FUNCTION_NAME;

	gtk_clist_append(GTK_CLIST(tree), data_string);
	for(n_columns-- ;n_columns>=0;n_columns--)
		gtk_clist_set_column_width(	GTK_CLIST(tree),n_columns,gtk_clist_optimal_column_width(GTK_CLIST(tree),n_columns));
}
#endif
#endif /* L_add_data_to_list */

#ifdef L_string_list_create_on_table
#ifdef USE_GTK_2_0
GtkWidget      *string_list_create_on_table(int num,	GtkSignalFunc func,
											GtkWidget *table, int start_col, int end_col,
											int start_row, int end_row ,int show_hide, ...)
{
	GtkCellRenderer		*renderer;
	GtkTreeViewColumn	*column;
	GtkListStore		*store;
	GtkWidget			*scrolled_window;
	GtkWidget			*tree;
	GtkTreeSelection *selection;
 	GType *types;
	va_list titles;
	int i;

	/* This is the scrolled window to put the cList widget inside */
	scrolled_window=gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),GTK_POLICY_ALWAYS,GTK_POLICY_ALWAYS);

	types = (GType *) malloc(num*sizeof(GType));
	for( i=0; i< num ; i++)
		types[i]=G_TYPE_STRING;

	store = gtk_list_store_newv (num, types);
	types=xfree(types);
	tree = gtk_tree_view_new_with_model (GTK_TREE_MODEL(store));

	/* Setup the selection handler */
	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (tree));
	gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);

	if(func!=NULL)
		signal_connect ( selection, "changed", G_CALLBACK (func), NULL);
	debug(if(func==NULL) err_msg("%s(): func == NULL",__FUNCTION__));

	gtk_container_add(GTK_CONTAINER(scrolled_window),GTK_WIDGET(tree));
	attach_to_table (table, GTK_WIDGET(scrolled_window), start_col, end_col, start_row, end_row);
	
	/* The view now holds a reference.  We can get rid of our own reference */
	g_object_unref (G_OBJECT (store));

	va_start(titles, show_hide);
	for(i=0; i<num ;i++)
	{
		/* Create a cell renderer */
		renderer = gtk_cell_renderer_text_new ();
		/* Create a column, associating the "text" attribute of the  cell_renderer to the column of the model */

		column = gtk_tree_view_column_new_with_attributes ((show_hide == SHOW) ? va_arg(titles, char *) : "" ,
								renderer,
								"text",
								i,
								NULL);
		/* Add the column to the view. */
		gtk_tree_view_append_column (GTK_TREE_VIEW (tree), column);
	}
	va_end(titles);
	gtk_scrolled_window_set_shadow_type (GTK_SCROLLED_WINDOW (scrolled_window), GTK_SHADOW_ETCHED_IN);
	return tree;
}
#else
GtkWidget  *string_list_create_on_table( int num, GtkSignalFunc func,
									 GtkWidget *table, int start_col, int end_col,
									 int start_row, int end_row ,int show_hide, ...)
{

	GtkWidget      *list;
	GtkWidget      *scrolled_window;
	va_list titles;

	debug(if(num<=0)err_msg("%s() num of columns <= 0",__FUNCTION__));
	
	/* This is the scrolled window to put the cList widget inside */
	scrolled_window=gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_window),GTK_POLICY_ALWAYS,GTK_POLICY_ALWAYS);
	attach_to_table (table, scrolled_window, start_col, end_col, start_row, end_row);

	va_start(titles, show_hide);
	list = gtk_clist_new_with_titles( num, (char **)titles);
	/* Set columns size to be at least as big as the title string */
	for(num--; num>=0 && titles!=NULL; num--, titles++)
		gtk_clist_set_column_min_width (GTK_CLIST(list), num, xstrlen(titles));
	va_end(titles);

	if(func!=NULL)
		signal_connect(GTK_OBJECT(list), "select_row", GTK_SIGNAL_FUNC(func), (gpointer) list);
	debug(if(func==NULL)err_msg("%s(): func == NULL",__FUNCTION__));

	/* It isn't necessary to shadow the border, but it looks nice :) */
    gtk_clist_set_shadow_type (GTK_CLIST(list), GTK_SHADOW_OUT);

	gtk_container_add(GTK_CONTAINER(scrolled_window), list);

	if(show_hide==HIDE)
		gtk_clist_column_titles_hide((GtkCList *) list);
	/* This is the default */
	/*else if(show_hide==SHOW)
		gtk_clist_column_titles_show((GtkCList *) list);*/
	
	return list;
}
#endif /* USE_GTK_2_0 */
#endif
/******** End of Create a Multi column scrolled list ************************/

/************** Create a statusbar **********************************/
#ifdef L_statusbar_create_on_table
GtkWidget * statusbar_create_on_table(GtkWidget *table, GtkWidget **label, char* labeltext, int start_col, int end_col, int start_row, int end_row )
{

	GtkWidget *frame;

	frame=frame_create_on_table(NULL,table, start_col, end_col, start_row, end_row );

	/* Allowed values for shadow type are:
		GTK_SHADOW_NONE,
		GTK_SHADOW_IN,
		GTK_SHADOW_OUT,
		GTK_SHADOW_ETCHED_IN,
		GTK_SHADOW_ETCHED_OUT
	*/

	gtk_frame_set_shadow_type(GTK_FRAME(frame),GTK_SHADOW_IN);
	*label=gtk_label_new(labeltext);
	gtk_misc_set_alignment((GtkMisc *)*label,  0.5, 0.5);
	gtk_container_add(GTK_CONTAINER(frame),GTK_WIDGET(*label));
	return frame;
}
#endif
/************** End of Create a statusbar **********************************/

/******************* Create a Textpad **************************************/
#ifdef L_textpad_create_on_table

#include <stdlib.h>

#ifdef USE_GTK_2_0
GtkWidget *textpad_create_on_table(int editable, GtkSignalFunc func, void * func_arg, int wrap, int scroll, GtkWidget *table,
										  int start_col,int end_col, int start_row, int end_row )
{
	GtkWidget		*text,
					*frame,
					*sw;

	text = gtk_text_view_new ();
	gtk_text_buffer_create_tag (gtk_text_view_get_buffer (GTK_TEXT_VIEW (text)), "monospace", "family", "monospace", NULL);
	frame = gtk_frame_new(NULL);
	attach_to_table (table, frame, start_col, end_col, start_row, end_row);
	gtk_text_view_set_editable (GTK_TEXT_VIEW(text), FALSE);
	if(editable==EDITABLE)
	{
		gtk_text_view_set_editable (GTK_TEXT_VIEW(text), TRUE);
		/* It's best to connect a signal here to have Gtk tell us if the text gets modified! */
		/* but only for text editing widgets so also NULL is allowed */
		if(func!=NULL)
			signal_connect(gtk_text_view_get_buffer (GTK_TEXT_VIEW (text)),"changed", GTK_SIGNAL_FUNC(func),func_arg);
		debug(if(func==NULL) err_msg("%s() EDITABLE but func == NULL",__FUNCTION__));
	}

	if(wrap==WRAP)
		gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW(text),GTK_WRAP_WORD);

	if(scroll==SCROLL)
	{
		sw = gtk_scrolled_window_new (NULL, NULL);
		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (sw),
						GTK_POLICY_AUTOMATIC,
						GTK_POLICY_ALWAYS);
		gtk_container_add (GTK_CONTAINER (frame), sw);
		gtk_container_add (GTK_CONTAINER (sw), text);
	}
	else
		gtk_container_add (GTK_CONTAINER (frame), text);
	
	return text;
}

#else

GtkWidget *textpad_create_on_table(int editable, GtkSignalFunc func, void * func_arg, int wrap, int scroll, GtkWidget *table,
										  int start_col,int end_col, int start_row, int end_row )
{
	GtkWidget		*new_text,
					*vscrollbar;

	new_text = gtk_text_new(NULL, NULL);
	gtk_text_set_editable (GTK_TEXT (new_text), FALSE);
	if(editable==EDITABLE)
	{
		gtk_text_set_editable (GTK_TEXT (new_text), TRUE);
		/* It's best to connect a signal here to have Gtk tell us if the text gets modified! */
		/* but only for text editing widgets so also NULL is allowed */
		if(func!=NULL)
			signal_connect (GTK_OBJECT (new_text),"changed", GTK_SIGNAL_FUNC(func),func_arg);
		debug(if(func==NULL) err_msg("%s() EDITABLE but GtkSignalFunc func == NULL",__FUNCTION__));
	}

	gtk_text_set_line_wrap(GTK_TEXT(new_text),(wrap==WRAP)?TRUE:FALSE);
	gtk_text_set_word_wrap(GTK_TEXT(new_text),(wrap==WRAP)?TRUE:FALSE);

	attach_to_table (table, new_text, start_col, end_col, start_row, end_row);
	if(scroll==SCROLL)
	{
		vscrollbar = gtk_vscrollbar_new(GTK_TEXT(new_text)->vadj);
		gtk_table_attach(GTK_TABLE(table), vscrollbar,	end_col, end_col+1, start_row, end_row,
							GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
	}

	return new_text;
}

#endif /* USE_GTK_2_0 */
#endif /*L_textpad_create_on_table*/


/************ End of Create a Textpad **************************************/

/************ Create a Main Menubar   **************************************/
#ifdef L_main_menu_create
#ifdef USE_GTK_2_4
GtkWidget * main_menu_create(GtkWidget *window, MenuEntry *menu_items, const gchar *info, gpointer data)
{
	GtkUIManager *ui;
	GtkActionGroup *actions;
	#ifdef DEBUG
	GError *error = NULL;
	#endif
	int i;

	actions = gtk_action_group_new ("Actions");
	
	for( i=0; ; i++) {
		if(menu_items[i].simple_entries!=NULL) {
			gtk_action_group_add_actions (actions,menu_items[i].simple_entries, menu_items[i].num_entries, data);
		} else if(menu_items[i].toggle_entries!=NULL) {
			gtk_action_group_add_toggle_actions (actions, menu_items[i].toggle_entries, menu_items[i].num_entries, data);
		} else if(menu_items[i].radio_entries!=NULL) {
			gtk_action_group_add_radio_actions (actions, menu_items[i].radio_entries,  menu_items[i].num_entries,
										menu_items[i].active_radio_entry,
										G_CALLBACK (menu_items[i].radio_entry_callback), data);
		} else {
			break;
		}
	}
	ui = gtk_ui_manager_new ();
	gtk_ui_manager_insert_action_group (ui, actions, 0);
	gtk_window_add_accel_group (GTK_WINDOW (window), gtk_ui_manager_get_accel_group (ui));
	

	#ifdef DEBUG
	if (gtk_ui_manager_add_ui_from_string (ui, info, -1, &error)==0) { /* something bad happened */
		g_message ("%s() building menus failed: %s", error->message,__FUNCTION__);
		g_error_free (error);
	}
	#else
	if (gtk_ui_manager_add_ui_from_string (ui, info, -1, NULL)==0) /* something bad happened */
		return NULL;
	#endif
	gtk_ui_manager_ensure_update (ui);  
	return gtk_ui_manager_get_widget (ui, "/MenuBar");	/* Returns NULL on failure */
}

#else
GtkWidget * main_menu_create( GtkWidget *window, GtkItemFactoryEntry *menu_items, gpointer ptr_action_num )
{
	GtkWidget *menubar=NULL;
	GtkItemFactory *item_factory;
	GtkAccelGroup *accel_group;
	int n;

	accel_group = gtk_accel_group_new ();

	/* This function initializes the item factory.
		Param 1: The type of menu - can be GTK_TYPE_MENU_BAR, GTK_TYPE_MENU,
				or GTK_TYPE_OPTION_MENU.
		Param 2: The path of the menu.
		Param 3: A pointer to a gtk_accel_group.  The item factory sets up
				the accelerator table while generating menus.
	*/
	item_factory = gtk_item_factory_new (GTK_TYPE_MENU_BAR, "<main>", accel_group);

	#ifdef USE_GTK_2_0
	g_object_set_data_full (G_OBJECT (window), "<main>", item_factory, (GDestroyNotify) g_object_unref);
	#endif

	for(n=0;;n++)
	{
		if(menu_items[n].path==NULL)
			break;
		/*	1 in arg 4 defines the type of callback
			(*GtkItemFactoryCallback1) (gpointer callback_data, guint callback_action, GtkWidget *widget);
			arg 3 is gpointer callback_data;
			guint callback_action  is field 4 of  GtkItemFactoryEntry text_edit_menu_items */
		gtk_item_factory_create_item(item_factory, &menu_items[n], ptr_action_num, 1);
	}

	/* Attach the new accelerator group to the window. */
	gtk_window_add_accel_group (GTK_WINDOW (window), accel_group);

	/* Finally, return the actual menu bar created by the item factory. */
	menubar = gtk_item_factory_get_widget (item_factory, "<main>");
	
	debug(if (menubar==NULL)err_msg("%s() menubar==NULL",__FUNCTION__));
	
	return menubar;
}
#endif
#endif /* L_main_menu_create */
/************ End of Create a Main Menubar**********************************/

/************ Progressbar widget  ******************************************/

#ifdef L_progress_timeout
#ifdef USE_GTK_2_0
gboolean progress_timeout( gpointer data )
{
	ProgressData *pdata = (ProgressData *)data;
	gdouble new_val=0.0;
	char percent[5];

	/* Update the value of the progress bar so that we get some movement */
	if (pdata->activity_mode==ACTIVITY) {
		/* Why do they callit pulse here?   */
		if(GTK_IS_PROGRESS_BAR (pdata->pbar))
			gtk_progress_bar_pulse (GTK_PROGRESS_BAR (pdata->pbar));
	} else {
		/* PULSE */
		/* Calculate the value of the progress bar using the
		* value range set in the adjustment object */
		if(pdata->use_value==USE_VALUE) {
			new_val=pdata->value;
		} else {
			if(GTK_IS_PROGRESS_BAR (pdata->pbar))
				new_val = gtk_progress_bar_get_fraction (GTK_PROGRESS_BAR (pdata->pbar)) + 0.01;
			else
				return FALSE;
		}
		/* Set the new value */
		snprintf(percent,5,"%.0f%%",(new_val*100));
		if(GTK_IS_PROGRESS_BAR (pdata->pbar))
			gtk_progress_bar_set_text (GTK_PROGRESS_BAR (pdata->pbar),  percent);
		if (new_val >= 0.00 && new_val <= 1.0) {
			if(GTK_IS_PROGRESS_BAR (pdata->pbar)) {
				gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (pdata->pbar), new_val);
			}
		} else {
			return FALSE;
		}
	}
	/* As this is a timeout function, return TRUE so that it
	* continues to get called */
	return TRUE;
} 

#else

gint progress_timeout( gpointer data )
{
	ProgressData *pdata = (ProgressData *)data;
	gfloat new_val;

	/* Calculate the value of the progress bar using the
	* value range set in the adjustment object */
	if(pdata->use_value==USE_VALUE && pdata->activity_mode==PULSE) {
		new_val=pdata->value;
		debug(err_msg("%f", new_val));
	} else {
		new_val = gtk_progress_get_value( GTK_PROGRESS(pdata->pbar) ) + 1;
	}
	
	if (new_val > pdata->adj->upper  && pdata->activity_mode==PULSE ) {
		debug(err_msg("return FALSE"));
		return (FALSE);
	}
	if(new_val > pdata->adj->upper  && pdata->activity_mode==ACTIVITY) {
		new_val = pdata->adj->lower;
		debug(err_msg("new_val > pdata->adj->upper"));
	}
	
	/* Set the new value */
	if(GTK_PROGRESS (pdata->pbar))
		gtk_progress_set_value (GTK_PROGRESS (pdata->pbar), new_val);
	/* As this is a timeout function, return TRUE so that it
	* continues to get called */
	return(TRUE);
} 
#endif 
#endif /* progress_timeout */

#ifdef L_progress_destroy
void progress_destroy(ProgressData *pdata)
{
	DBG_FUNCTION_NAME;

	debug(err_msg("pdata is%s NULL",(pdata==NULL)?"":" not"));	
	debug(err_msg("GTK_IS_WIDGET(pdata->pbar)==%s",(GTK_IS_WIDGET(pdata->pbar))?"TRUE":"FALSE"));
	debug(err_msg("pdata->timer is %d", pdata->timer));
	debug(err_msg("pdata->alive is %d", pdata->alive));
	
	if(pdata==NULL){
		return;
	}
	if(GTK_IS_WIDGET(pdata->pbar)){
		gtk_widget_hide(pdata->pbar);
	}
	if(pdata->timer >= 0) {
		timeout_remove (pdata->timer);
	}
	pdata->timer = 0;
	pdata->value = 0;
	if(pdata->alive == 0 && GTK_IS_WIDGET(pdata->window)) {
		gtk_widget_destroy(pdata->window);
		pdata->window=NULL;
	}
	if(GTK_IS_WIDGET(pdata->pbar)){
		gtk_widget_destroy(pdata->pbar);
		pdata->pbar=NULL;
	}
	pdata=xfree(pdata);

}
#endif

#ifdef L_progressbar_create_on_table
ProgressData *progressbar_create_on_table(int mode, int use_value, GtkWidget *table,
											int start_col, int end_col, int start_row, int end_row )
{
	ProgressData *pdata=progressbar_create(mode, use_value);
	attach_to_table (table, pdata->pbar, start_col, end_col, start_row, end_row);
	return pdata;
}
#endif

#ifdef L_progressbar_create
ProgressData *progressbar_create(int mode, int use_value)
{
	ProgressData *pdata;
	#ifndef USE_GTK_2_0
	GtkAdjustment *adj;
	#endif
	
	/* Allocate memory for the data that is passed to the callbacks */
	pdata = xmalloc (sizeof (ProgressData));
	
	/* Initialize */
	pdata->window=NULL;
	/* destroy when done */
	pdata->alive=0;
	
	#ifdef USE_GTK_2_0
	/* Create the GtkProgressBar */
	pdata->pbar = gtk_progress_bar_new ();
	#else
	/* Create a Adjusment object to hold the range of the progress bar */
	adj = (GtkAdjustment *) gtk_adjustment_new (0.01, 0.0, 1.0, 0, 0, 0);
	/* Create the GtkProgressBar using the adjustment */
	pdata->pbar = gtk_progress_bar_new_with_adjustment (adj);
	pdata->adj=adj;
	#endif
	
	/* Add a timer callback to update the value of the progress bar */
	pdata->timer = timeout_add (100, progress_timeout, pdata);
	pdata->activity_mode=mode;
	pdata->value=0.0;

	debug(if(pdata->activity_mode==ACTIVITY && use_value==1)
				err_msg("%s: ACTIVITY mode: use_value=1 is ignored",__FUNCTION__));
	
	if(pdata->activity_mode==PULSE && use_value==USE_VALUE) {
		pdata->use_value=USE_VALUE;
 	} else {
		pdata->use_value=NOT_USE_VALUE;
	}
	#ifndef USE_GTK_2_0
		gtk_progress_set_activity_mode (GTK_PROGRESS (pdata->pbar),(pdata->activity_mode==ACTIVITY) ? TRUE : FALSE);
		
	if(pdata->activity_mode==PULSE) {
		gtk_progress_set_format_string (GTK_PROGRESS(pdata->pbar),"%p%%");
		gtk_progress_set_show_text (GTK_PROGRESS (pdata->pbar),TRUE); 
	}
	#endif

	return pdata;
}
#endif

/************ Progressbar widget  ******************************************/

/************ standard about window widget  ********************************/

#ifdef L_about_callback
void about_callback(void)
{
	DBG_FUNCTION_NAME;
	dialog_window(STR_ABOUT, ONE_BUTTON, NULL, 1, VERSION"\n"COPYRIGHT"\n"STR_LICENSE"\n"HOMEPAGE"\n"STR_WARRANTY);
}
#endif

/************ standard about window widget  ********************************/
